@@ -2,14 +2,14 @@
  ** osys.c
  ** System extensions.
  **
- ** Copyright (C) 2003-2007 Cosmin Truta.
+ ** Copyright (C) 2003-2008 Cosmin Truta.
  **
  ** This software is distributed under the same licensing and warranty
  ** terms as OptiPNG.  Please see the attached LICENSE for more info.
  **/
 
 
-#if defined UNIX || defined unix
+#if defined UNIX || defined __unix || defined __unix__
 # define OSYS_UNIX
 #endif
 
@@ -24,7 +24,7 @@
 # endif
 #endif
 
-#if defined MSDOS || defined __MSDOS__
+#if defined MSDOS || defined _MSDOS || defined __MSDOS__
 # define OSYS_DOS
 #endif
 
@@ -90,11 +90,33 @@
     defined OSYS_WINDOWS || defined __CYGWIN__
 # define OSYS_FNAME_DOS
 # define OSYS_FNAME_IGN_CASE 1
-#else  /* OSYS_UNIX and possibly others */
+#else  /* OSYS_UNIX and others */
 # define OSYS_FNAME_IGN_CASE 0
 #endif
 
 
+#ifdef R_OK
+# define OSYS_FTEST_READ R_OK
+#else
+# define OSYS_FTEST_READ 4
+#endif
+#ifdef W_OK
+# define OSYS_FTEST_WRITE W_OK
+#else
+# define OSYS_FTEST_WRITE 2
+#endif
+#ifdef X_OK
+# define OSYS_FTEST_EXEC X_OK
+#else
+# define OSYS_FTEST_EXEC 1
+#endif
+#ifdef F_OK
+# define OSYS_FTEST_FILE F_OK
+#else
+# define OSYS_FTEST_FILE 0
+#endif
+
+
 /**
  * Allocates memory safely.
  * On success, the function returns the address of the allocated block.
@@ -150,7 +172,7 @@ void osys_terminate(void)
 
     ExitProcess(EXIT_FAILURE);
 
-#else
+#else  /* generic */
 
     exit(EXIT_FAILURE);
 
@@ -160,8 +182,8 @@ void osys_terminate(void)
 
 /**
  * Creates a backup file name.
- * On success, returns buffer.
- * On error, returns NULL.
+ * On success, the function returns buffer.
+ * On error, it returns NULL.
  **/
 char *osys_fname_mkbak(char *buffer, size_t bufsize, const char *fname)
 {
@@ -172,7 +194,7 @@ char *osys_fname_mkbak(char *buffer, size_t bufsize, const char *fname)
 
     return osys_fname_chext(buffer, bufsize, fname, OSYS_FNAME_STR_DOT "bak");
 
-#else
+#else  /* OSYS_UNIX and others */
 
     strcpy(buffer, fname);
     strcat(buffer, OSYS_FNAME_STR_DOT "bak");
@@ -188,17 +210,17 @@ char *osys_fname_mkbak(char *buffer, size_t bufsize, const char *fname)
  * the new file name has no directory (or is in the default directory).
  * The directory name may or may not contain the trailing directory
  * separator (usually '/').
- * On success, returns buffer.
- * On error, returns NULL.
+ * On success, the function returns buffer.
+ * On error, it returns NULL.
  **/
 char *osys_fname_chdir(char *buffer, size_t bufsize,
-    const char *oldname, const char *newdir)
+    const char *old_fname, const char *new_dirname)
 {
     const char *fname, *ptr;
     size_t dirlen;
 
-    /* Extract file name from oldname. */
-    fname = oldname;
+    /* Extract file name from old_fname. */
+    fname = old_fname;
 #ifdef OSYS_FNAME_DOS
     if (isalpha(fname[0]) && fname[1] == ':')
         fname += 2;  /* skip drive name */
@@ -212,14 +234,14 @@ char *osys_fname_chdir(char *buffer, size_t bufsize,
     }
 
     /* Make sure the buffer is large enough. */
-    dirlen = strlen(newdir);
+    dirlen = strlen(new_dirname);
     if (dirlen + strlen(fname) + 2 >= bufsize)
         return NULL;  /* overflow */
 
     /* Copy the new directory name. Also append a slash if necessary. */
     if (dirlen > 0)
     {
-        strcpy(buffer, newdir);
+        strcpy(buffer, new_dirname);
 #ifdef OSYS_FNAME_DOS
         if (dirlen == 2 && buffer[1] == ':' && isalpha(buffer[0]))
             (void)0;  /* do nothing */
@@ -240,30 +262,30 @@ char *osys_fname_chdir(char *buffer, size_t bufsize,
  * The new extension can be the empty string, indicating that the new
  * file name has no extension.  Otherwise, it must begin with the
  * extension separator (usually '.').
- * On success, returns buffer.
- * On error, returns NULL.
+ * On success, the function returns buffer.
+ * On error, it returns NULL.
  **/
 char *osys_fname_chext(char *buffer, size_t bufsize,
-    const char *oldname, const char *newext)
+    const char *old_fname, const char *new_extname)
 {
     size_t i, pos;
 
-    if (newext[0] != OSYS_FNAME_CHR_DOT)
+    if (new_extname[0] != OSYS_FNAME_CHR_DOT)
         return NULL;  /* invalid argument */
-    for (i = 0, pos = (size_t)(-1); oldname[i] != 0; ++i)
+    for (i = 0, pos = (size_t)(-1); old_fname[i] != 0; ++i)
     {
         if (i >= bufsize)
             return NULL;  /* overflow */
-        if ((buffer[i] = oldname[i]) == OSYS_FNAME_CHR_DOT)
+        if ((buffer[i] = old_fname[i]) == OSYS_FNAME_CHR_DOT)
             pos = i;
     }
     if (i > pos)
-        i = pos;  /* go back only if oldname has an extension */
-    for ( ;; ++i, ++newext)
+        i = pos;  /* go back only if old_fname has an extension */
+    for ( ; ; ++i, ++new_extname)
     {
         if (i >= bufsize)
             return NULL;  /* overflow */
-        if ((buffer[i] = *newext) == 0)
+        if ((buffer[i] = *new_extname) == 0)
             return buffer;  /* success */
     }
 }
@@ -272,24 +294,47 @@ char *osys_fname_chext(char *buffer, size_t bufsize,
 /**
  * Compares one file name to another.
  * It returns a value (less than, equal to, or greater than 0)
- * based on the result of comparing name1 to name2.
+ * based on the result of comparing fname1 to fname2.
  * The comparison may or may not be case sensitive, depending on
  * the operating system.
  **/
-int osys_fname_cmp(const char *name1, const char *name2)
+int osys_fname_cmp(const char *fname1, const char *fname2)
 {
 #if OSYS_FNAME_IGN_CASE
 # ifdef OSYS_WINDOWS
-    return lstrcmpiA(name1, name2);
+    return lstrcmpiA(fname1, fname2);
 # else
-    return stricmp(name1, name2);
+    return stricmp(fname1, fname2);
 # endif
 #else
-    return strcmp(name1, name2);
+    return strcmp(fname1, fname2);
 #endif
 }
 
 
+/**
+ * Opens a file and positions it at the specified file offset.
+ * On success, the function returns the pointer to the file stream.
+ * On error, it returns NULL.
+ **/
+FILE *osys_fopen_at(const char *fname, const char *mode,
+    long offset, int whence)
+{
+    FILE *stream;
+
+    if ((stream = fopen(fname, mode)) == NULL)
+        return NULL;
+    if (offset == 0 && (whence == SEEK_SET || whence == SEEK_CUR))
+        return stream;
+    if (fseek(stream, offset, whence) != 0)
+    {
+        fclose(stream);
+        return NULL;
+    }
+    return stream;
+}
+
+
 /**
  * Reads a block of data from the specified file offset.
  * The file-position indicator is saved and restored after reading.
@@ -303,16 +348,16 @@ size_t osys_fread_at(FILE *stream, long offset, int whence,
     fpos_t pos;
     size_t result;
 
-    if (fflush(stream) != 0 || block == NULL || blocksize == 0
-            || fgetpos(stream, &pos) != 0)
+    if (fgetpos(stream, &pos) != 0 || fflush(stream) != 0)
         return 0;
     if (fseek(stream, offset, whence) == 0)
         result = fread(block, 1, blocksize, stream);
     else
         result = 0;
-    if (fflush(stream) != 0) result = 0;
-    if (fsetpos(stream, &pos) != 0) result = 0;
-    if (fflush(stream) != 0) result = 0;
+    if (fflush(stream) != 0)
+        result = 0;
+    if (fsetpos(stream, &pos) != 0)
+        result = 0;
     return result;
 }
 
@@ -330,27 +375,106 @@ size_t osys_fwrite_at(FILE *stream, long offset, int whence,
     fpos_t pos;
     size_t result;
 
-    if (fflush(stream) != 0 || block == NULL || blocksize == 0
-            || fgetpos(stream, &pos) != 0)
+    if (fgetpos(stream, &pos) != 0 || fflush(stream) != 0)
         return 0;
     if (fseek(stream, offset, whence) == 0)
         result = fwrite(block, 1, blocksize, stream);
     else
         result = 0;
-    if (fflush(stream) != 0) result = 0;
-    if (fsetpos(stream, &pos) != 0) result = 0;
-    if (fflush(stream) != 0) result = 0;
+    if (fflush(stream) != 0)
+        result = 0;
+    if (fsetpos(stream, &pos) != 0)
+        result = 0;
     return result;
 }
 
 
 /**
- * Copies the file mode and the time stamp of the file
- * named by destname into the file named by srcname.
- * On success, returns 0.
- * On error, sets the global variable errno and returns -1.
+ * Determines if the accessibility of the specified file satisfies
+ * the specified access mode.  The access mode consists of one or more
+ * characters that indicate the checks to be performed, as follows:
+ *  'e': the file exists; it needs not be a regular file.
+ *  'f': the file exists and is a regular file.
+ *  'r': the file exists and read permission is granted.
+ *  'w': the file exists and write permission is granted.
+ *  'x': the file exists and execute permission is granted.
+ * For example, to determine if a file can be opened for reading using
+ * fopen(), use "fr" in the access mode.
+ * If all checks succeed, the function returns 0.
+ * Otherwise, it returns -1.
  **/
-int osys_fattr_copy(const char *destname, const char *srcname)
+int osys_ftest(const char *fname, const char *mode)
+{
+    int faccess, freg;
+
+    faccess = freg = 0;
+    if (strchr(mode, 'f') != NULL)
+        freg = 1;
+    if (strchr(mode, 'r') != NULL)
+        faccess |= OSYS_FTEST_READ;
+    if (strchr(mode, 'w') != NULL)
+        faccess |= OSYS_FTEST_WRITE;
+    if (strchr(mode, 'x') != NULL)
+        faccess |= OSYS_FTEST_EXEC;
+    if (faccess == 0 && freg == 0)
+        if (strchr(mode, 'e') == NULL)
+            return 0;
+
+#if defined OSYS_UNIX || defined OSYS_DOS || defined OSYS_OS2
+
+    {
+        struct stat sbuf;
+
+        if (stat(fname, &sbuf) != 0)
+            return -1;
+        if (freg && !(sbuf.st_mode & S_IFREG))
+            return -1;
+        if (faccess == 0)
+            return 0;
+        return access(fname, faccess);
+    }
+
+#elif defined OSYS_WINDOWS
+
+    {
+        DWORD attr;
+
+        attr = GetFileAttributesA(fname);
+        if (attr == 0xffffffffU)
+            return -1;
+        if (freg && (attr & FILE_ATTRIBUTE_DIRECTORY))
+            return -1;
+        if ((faccess & OSYS_FTEST_WRITE) && (attr & FILE_ATTRIBUTE_READONLY))
+            return -1;
+        return 0;
+    }
+
+#else  /* generic */
+
+    {
+        FILE *stream;
+
+        if (faccess & OSYS_FTEST_WRITE)
+            stream = fopen(fname, "r+b");
+        else
+            stream = fopen(fname, "rb");
+        if (stream == NULL)
+            return -1;
+        fclose(stream);
+        return 0;
+    }
+
+#endif
+}
+
+
+/**
+ * Copies the access mode and the time stamp of the file or directory
+ * named by dest_name into the file or directory named by src_name.
+ * On success, the function returns 0.
+ * On error, it returns -1.
+ **/
+int osys_fattr_copy(const char *dest_name, const char *src_name)
 {
 #if defined OSYS_UNIX || defined OSYS_DOS || defined OSYS_OS2
 
@@ -358,25 +482,29 @@ int osys_fattr_copy(const char *destname, const char *srcname)
     int /* mode_t */ mode;
     struct utimbuf utbuf;
 
-    if (stat(srcname, &sbuf) != 0)
+    if (stat(src_name, &sbuf) != 0)
         return -1;
 
     mode = (int)sbuf.st_mode;
     utbuf.actime = sbuf.st_atime;
     utbuf.modtime = sbuf.st_mtime;
 
-    if (utime(destname, &utbuf) == 0 && chmod(destname, mode) == 0)
+    if (utime(dest_name, &utbuf) == 0 && chmod(dest_name, mode) == 0)
         return 0;
     else
         return -1;
 
 #elif defined OSYS_WINDOWS
 
+    static int isWinNT = -1;
     HANDLE hFile;
     FILETIME ftLastWrite;
     BOOL result;
 
-    hFile = CreateFileA(srcname, GENERIC_READ, 0, NULL, OPEN_EXISTING, 0, 0);
+    if (isWinNT < 0)
+        isWinNT = (GetVersion() < 0x80000000U) ? 1 : 0;
+
+    hFile = CreateFileA(src_name, GENERIC_READ, 0, NULL, OPEN_EXISTING, 0, 0);
     if (hFile == INVALID_HANDLE_VALUE)
         return -1;
     result = GetFileTime(hFile, NULL, NULL, &ftLastWrite);
@@ -384,7 +512,8 @@ int osys_fattr_copy(const char *destname, const char *srcname)
     if (!result)
         return -1;
 
-    hFile = CreateFileA(destname, GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 0, 0);
+    hFile = CreateFileA(dest_name, GENERIC_WRITE, 0, NULL, OPEN_EXISTING,
+        (isWinNT ? FILE_FLAG_BACKUP_SEMANTICS : 0), 0);
     if (hFile == INVALID_HANDLE_VALUE)
         return -1;
     result = SetFileTime(hFile, NULL, NULL, &ftLastWrite);
@@ -392,11 +521,11 @@ int osys_fattr_copy(const char *destname, const char *srcname)
     if (!result)
         return -1;
 
+    /* TODO: Copy the access mode. */
+
     return 0;
 
-    /* TODO: Copy file access mode. */
-
-#else
+#else  /* generic */
 
     /* Do nothing. */
     return 0;
@@ -409,7 +538,7 @@ int osys_fattr_copy(const char *destname, const char *srcname)
  * Creates a new directory with the given name.
  * If the directory is successfully created, or if it already exists,
  * the function returns 0.
- * Otherwise, it sets the global variable errno and returns -1.
+ * Otherwise, it returns -1.
  **/
 int osys_dir_make(const char *dirname)
 {
@@ -427,13 +556,10 @@ int osys_dir_make(const char *dirname)
 #if defined OSYS_UNIX || defined OSYS_DOS || defined OSYS_OS2
 
     {
-        DIR *dir;
+        struct stat sbuf;
 
-        if ((dir = opendir(dirname)) != NULL)
-        {
-            closedir(dir);
-            return 0;
-        }
+        if (stat(dirname, &sbuf) == 0)
+            return (sbuf.st_mode & S_IFDIR) ? 0 : -1;
 
         /* There is no directory, so create one now. */
 # if defined OSYS_DOS || defined OSYS_OS2
@@ -473,7 +599,7 @@ int osys_dir_make(const char *dirname)
         return CreateDirectoryA(dirname, NULL) ? 0 : -1;
     }
 
-#else
+#else  /* generic */
 
     /* Do nothing. */
     return 0;
